import { Injectable } from '@angular/core';
import { HttpClient, HttpResponse } from '@angular/common/http';
import { Observable } from 'rxjs';
import { map } from 'rxjs/operators';
import moment from 'moment';

import { DATE_FORMAT } from 'app/config/input.constants';
import { ApplicationConfigService } from 'app/core/config/application-config.service';
import { createRequestOption } from 'app/core/request/request-util';
import { Search } from 'app/core/request/request.model';
import { ILikes, Likes } from 'app/shared/model/likes.model';
import { encodeURIforExerciseId } from 'app/exercise/service/exercise.service'
type EntityResponseType = HttpResponse<ILikes>;
type EntityArrayResponseType = HttpResponse<ILikes[]>;

@Injectable({ providedIn: 'root' })
export class LikesService {
  public resourceUrl = this.applicationConfigService.getEndpointFor('api/likes');
  public resourceSearchUrl = this.applicationConfigService.getEndpointFor('api/_search/likes');

  constructor(protected http: HttpClient, private applicationConfigService: ApplicationConfigService) {}

  create(likes: ILikes): Observable<EntityResponseType> {
    const copy = this.convertDateFromClient(likes);
    return this.http
      .post<ILikes>(this.resourceUrl, copy, { observe: 'response' })
      .pipe(map((res: EntityResponseType) => this.convertDateFromServer(res)));
  }

  update(likes: ILikes): Observable<EntityResponseType> {
    const copy = this.convertDateFromClient(likes);
    return this.http
      .put<ILikes>(this.resourceUrl, copy, { observe: 'response' })
      .pipe(map((res: EntityResponseType) => this.convertDateFromServer(res)));
  }

  find(id: number): Observable<EntityResponseType> {
    return this.http
      .get<ILikes>(`${this.resourceUrl}/${encodeURIComponent(id)}`, { observe: 'response' })
      .pipe(map((res: EntityResponseType) => this.convertDateFromServer(res)));
  }

  query(req?: any): Observable<EntityArrayResponseType> {
    const options = createRequestOption(req);
    return this.http
      .get<ILikes[]>(this.resourceUrl, { params: options, observe: 'response' })
      .pipe(map((res: EntityArrayResponseType) => this.convertDateArrayFromServer(res)));
  }

  delete(id: number): Observable<HttpResponse<{}>> {
    return this.http.delete(`${this.resourceUrl}/${id}`, { observe: 'response' });
  }

  search(req: Search): Observable<EntityArrayResponseType> {
    const options = createRequestOption(req);
    return this.http
      .get<ILikes[]>(this.resourceSearchUrl, { params: options, observe: 'response' })
      .pipe(map((res: EntityArrayResponseType) => this.convertDateArrayFromServer(res)));
  }

  likeExercise(exerciseID: string): Observable<EntityResponseType> {
    const like = new Likes();
    like.exerciseID = exerciseID;
    return this.http.put<ILikes>(SERVER_API_URL + 'api/likes/likeExercise', like, { observe: 'response' });
  }

  unlikeExercise(exerciseID: string): Observable<HttpResponse<{}>> {
    return this.http.delete(SERVER_API_URL + 'api/likes/unlikeExercise/' + encodeURIforExerciseId(exerciseID), { observe: 'response' });
  }

  getLikesForexerciseID(exerciseID: string): Observable<number> {
    return this.http.get<number>(SERVER_API_URL + 'api/likes/numberOfLikes/' + encodeURIforExerciseId(exerciseID));
  }

  hasLiked(exerciseID: string): Observable<boolean> {
    return this.http.get<boolean>(SERVER_API_URL + 'api/likes/hasLiked/' + encodeURIforExerciseId(exerciseID));
  }

  protected convertDateFromClient(likes: ILikes): ILikes {
    const copy: ILikes = Object.assign({}, likes, {
      date: likes.date && likes.date.isValid() ? likes.date.format(DATE_FORMAT) : undefined,
    });
    return copy;
  }

  protected convertDateFromServer(res: EntityResponseType): EntityResponseType {
    if (res.body) {
      res.body.date = res.body.date ? moment(res.body.date) : undefined;
    }
    return res;
  }

  protected convertDateArrayFromServer(res: EntityArrayResponseType): EntityArrayResponseType {
    if (res.body) {
      res.body.forEach((likes: ILikes) => {
        likes.date = likes.date ? moment(likes.date) : undefined;
      });
    }
    return res;
  }
}
