package at.ac.uibk.gitsearch.service;

import static org.hamcrest.CoreMatchers.containsStringIgnoringCase;
import static org.hamcrest.Matchers.anyOf;
import static org.hamcrest.Matchers.contains;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.everyItem;
import static org.hamcrest.Matchers.hasItemInArray;
import static org.hamcrest.Matchers.hasProperty;
import static org.hamcrest.Matchers.is;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import at.ac.uibk.gitsearch.IntegrationTest;
import at.ac.uibk.gitsearch.repository.gitlab.GitLabRepository;
import at.ac.uibk.gitsearch.repository.search.testESService.ElasticSearchTestConfiguration;
import at.ac.uibk.gitsearch.service.PluginManagementService.ConnectorConfigWrapper;
import at.ac.uibk.gitsearch.service.dto.AutoCompleteEntry;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.TimeUnit;
import javax.ws.rs.NotFoundException;
import org.codeability.sharing.plugins.api.SharingPluginConfig;
import org.codeability.sharing.plugins.api.search.SearchInputDTO;
import org.codeability.sharing.plugins.api.search.SearchInputMetadataDTO;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.SearchResultsDTO;
import org.elasticsearch.client.RestHighLevelClient;
import org.elasticsearch.node.NodeValidationException;
import org.gitlab4j.api.GitLabApi;
import org.gitlab4j.api.GitLabApiException;
import org.gitlab4j.api.RepositoryApi;
import org.hamcrest.MatcherAssert;
import org.junit.Assert;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.Timeout;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.security.test.context.support.WithAnonymousUser;
import org.springframework.security.test.context.support.WithMockUser;

@IntegrationTest
@WithMockUser(value = "test", authorities = "sharing")
public class SearchServiceIT {

    @Autowired
    RestHighLevelClient searchClient;

    @MockBean
    PluginManagementService pluginManagementService;

    @MockBean
    GitLabRepository gitLabRepository;

    @MockBean
    RepositoryApi mockedRepositoryApi;

    private static final Logger LOGGER = LoggerFactory.getLogger(SearchServiceIT.class);

    @Autowired
    private SearchService searchService;

    @Autowired
    ElasticSearchTestConfiguration elasticSearchTestConfiguration;

    static ElasticSearchTestConfiguration staticElasticSearchTestConfiguration;

    @BeforeEach // must be started as BeforeEach, in order to autowire the ElasticSearchTestConfiguration
    @Timeout(value = 2, unit = TimeUnit.MINUTES)
    public void setUpESServer() throws IOException, NodeValidationException {
        if (staticElasticSearchTestConfiguration == null) {
            elasticSearchTestConfiguration.startTestNode();
            staticElasticSearchTestConfiguration = elasticSearchTestConfiguration;
        }
    }

    @AfterAll
    public static void shutDownESServer() {
        if (staticElasticSearchTestConfiguration != null) {
            staticElasticSearchTestConfiguration.stopTestContainer();
            staticElasticSearchTestConfiguration = null;
        }
    }

    @BeforeEach
    public void initMock() {
        Collection<ConnectorConfigWrapper> mockConfigs = new ArrayList<>();
        mockConfigs.add(
            new ConnectorConfigWrapper(
                new SharingPluginConfig(
                    "mockedPlugin",
                    new SharingPluginConfig.Action[] {
                        new SharingPluginConfig.Action("mockedAction", "http://unused/unused-transferURL", "mocked Action", "true"),
                    }
                ),
                "http://unused/hopefully-unused"
            )
        );

        mockConfigs.add(
            new ConnectorConfigWrapper(
                new SharingPluginConfig(
                    "Artemis Sharing Connector",
                    new SharingPluginConfig.Action[] {
                        new SharingPluginConfig.Action("Import", "http://unused/unused-transferURL", "Export to Artemis", "true"),
                    }
                ),
                "http://unused/hopefully-unused"
            )
        );

        when(pluginManagementService.getRegisteredPluginConfigs()).thenReturn(mockConfigs);
    }

    @Test
    public void testAllSearch() throws Exception {
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, null, null);
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

        Assert.assertTrue("At least one hit?", searchResultPage.getSearchResult().size() >= 1);
        Assert.assertEquals("We start at 0", 0, searchResultPage.getPageStartIndex());
        LOGGER.info("found {} hits for all", searchResultPage.getHitCount());
    }

    @Test
    public void testFullTextSearch() throws Exception {
        LOGGER.info("starting testFullTextSearch");
        final String wien = "Wien";
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, null, null);
        SearchInputDTO searchQuery = new SearchInputDTO(wien, searchMetadata, null, null, null, 0);
        SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

        org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
        org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
        MatcherAssert.assertThat(
            searchResultPage.getSearchResult(),
            everyItem(
                anyOf(
                    hasProperty("metadata", hasProperty("description", containsString(wien))),
                    hasProperty("metadata", hasProperty("title", containsString(wien)))
                    // hasProperty("keywords", containsString("Wien"))
                )
            )
        );

        LOGGER.info("exiting testFullTextSearch");
    }

    @Test
    public void testSearchByAutor() throws Exception {
        final String podlipnig = "Stefan Podlipnig";
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, podlipnig, null);
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

        org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
        org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
        MatcherAssert.assertThat(
            searchResultPage.getSearchResult(),
            everyItem(hasProperty("metadata", hasProperty("creator", hasItemInArray(hasProperty("name", containsString(podlipnig))))))
        );
    }

    @Test
    @Disabled("Ich brings nicht zum laufen :-(")
    public void testProgrammingLanguageSearch() throws Exception {
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO("Java", null, null, null, null, null);
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

        assertNotNull(searchResultPage.getSearchResult());
        assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
        MatcherAssert.assertThat(
            searchResultPage.getSearchResult(),
            everyItem(hasProperty("metadata", hasProperty("programmingLanguage", hasItemInArray(containsStringIgnoringCase("Java")))))
        );
    }

    @Test
    public void testCreatorSearch() throws IOException {
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, "Podlipnig", null);
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

        org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
        org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
        MatcherAssert.assertThat(
            searchResultPage.getSearchResult(),
            everyItem(hasProperty("metadata", hasProperty("creator", hasItemInArray(hasProperty("name", containsString("Podlipnig"))))))
        );
    }

    @Test
    public void testKeywordSearch() throws IOException {
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, "latex", null, null, null, null);
        org.codeability.sharing.plugins.api.search.SearchInputDTO searchQuery = new SearchInputDTO(
            null,
            searchMetadata,
            null,
            null,
            null,
            0
        );
        SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

        org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
        org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
        MatcherAssert.assertThat(
            searchResultPage.getSearchResult(),
            everyItem(hasProperty("metadata", hasProperty("keyword", hasItemInArray(containsString("latex")))))
        );
    }

    @Test
    public void testFormatSearch() throws IOException {
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, null, "latex");
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

        org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
        org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
        MatcherAssert.assertThat(
            searchResultPage.getSearchResult(),
            everyItem(hasProperty("metadata", hasProperty("format", hasItemInArray(containsString("latex")))))
        );
    }

    // does not work with metadata version 0.4
    //	@Test
    //	public void testTypesSearch() throws IOException {
    //		final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, "latex", null, null, null,null);
    //		searchMetadata.setTypes(Collections.singletonList(ExerciseType.COLLECTION));
    //		SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
    //		SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);
    //
    //		org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
    //		org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
    //		assertThat(searchResultPage.getSearchResult(),
    //				everyItem(hasProperty("metadata", hasProperty("keyword", hasItemInArray(containsString("latex"))))));
    //
    //	}

    @Test
    public void testLanguagesSearch() throws IOException {
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, "latex", null, null, null, null);
        searchMetadata.setNaturalLanguage(List.of("de"));
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);
        org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());

        org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
        MatcherAssert.assertThat(
            searchResultPage.getSearchResult(),
            everyItem(hasProperty("metadata", hasProperty("language", hasItemInArray(containsString("de")))))
        );
    }

    // does not work with metadata version 0.4
    //	@Test
    //	public void testTypesSearchNegative() throws IOException {
    //		final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, "latex", null, null, null,null);
    //		searchMetadata.setTypes(Collections.singletonList(ExerciseType.OTHER));
    //		SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
    //		SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);
    //
    //		org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
    //		org.junit.Assert.assertTrue("We expect no hit for \"latex\" and ExerciseType.OTHER", searchResultPage.getHitCount() == 0);
    //
    //	}

    //	@Test()
    //	@Ignore() // Test funktioniert momentan nicht?
    public void testLicenseSearch() throws IOException {
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, "MIT", null, null);
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);

        org.junit.Assert.assertNotNull(searchResultPage.getSearchResult());
        org.junit.Assert.assertTrue("At least one test hit", searchResultPage.getHitCount() >= 1);
        MatcherAssert.assertThat(
            searchResultPage.getSearchResult(),
            everyItem(hasProperty("metadata", hasProperty("license", containsString("MIT"))))
        );
    }

    @Test
    public void testKeywordAutocompletion() throws IOException {
        final List<AutoCompleteEntry> keywordsAutoComplete = searchService.getKeywordsAutoComplete("Jav", 10);
        MatcherAssert.assertThat(keywordsAutoComplete, contains(hasProperty("target", is("Java"))));
    }

    @Test
    public void testCreatorAutocompletion() throws IOException {
        final List<AutoCompleteEntry> creatorAutoComplete = searchService.getCreatorAutoComplete("Pod", 10);
        MatcherAssert.assertThat(creatorAutoComplete, contains(hasProperty("target", is("Stefan Podlipnig"))));
        final List<AutoCompleteEntry> creatorAutoComplete2 = searchService.getCreatorAutoComplete("Po", 10);
        MatcherAssert.assertThat(creatorAutoComplete2, contains(hasProperty("target", is("Stefan Podlipnig"))));
    }

    @Test
    public void testContributorAutocompletion() throws IOException {
        final List<AutoCompleteEntry> contributorAutoComplete = searchService.getContributorAutoComplete("Bast", 10);
        MatcherAssert.assertThat(contributorAutoComplete, contains(hasProperty("target", is("Daniel Bastta"))));
    }

    @Test
    public void testContributorCreatorAutocompletion() throws IOException {
        final List<AutoCompleteEntry> contributorAutoComplete = searchService.getContributorCreatorAutoComplete("Bast", 10);
        MatcherAssert.assertThat(contributorAutoComplete, contains(hasProperty("target", is("Daniel Bastta"))));
    }

    @SuppressWarnings("unchecked")
    @Test
    public void testArtemisExportActionFilteringForAuthorizedUser() throws IOException, GitLabApiException {
        GitLabApi mockedGitLabApi = mock(GitLabApi.class);
        when(gitLabRepository.getGitLabApi(any(Optional.class))).thenReturn(mockedGitLabApi);
        doReturn(mockedRepositoryApi).when(mockedGitLabApi).getRepositoryApi();
        doReturn(new ByteArrayInputStream(" ".getBytes()))
            .when(mockedRepositoryApi)
            .getRepositoryArchive(eq(333), anyString(), anyString());

        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, null, null);
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        SearchResultsDTO results = searchService.searchResultPage(searchQuery);

        Optional<SearchResultDTO> result = results
            .getSearchResult()
            .stream()
            .filter(entry -> entry.getExerciseId().equals("[333]"))
            .findFirst();

        if (result.isEmpty()) {
            throw new NotFoundException("Exercise with publicVisibility and id '333' couldn't be found in the test data");
        }

        Assertions.assertTrue(
            result.get().getSupportedActions().stream().anyMatch(action -> action.getPlugin().equals("Artemis Sharing Connector"))
        );
    }

    @Test
    @WithMockUser(value = "User without authorities")
    public void testArtemisExportActionFilteringForUnauthorizedUser() throws IOException {
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, null, null);
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        SearchResultsDTO results = searchService.searchResultPage(searchQuery);

        Optional<SearchResultDTO> result = results
            .getSearchResult()
            .stream()
            .filter(entry -> entry.getExerciseId().equals("[333]"))
            .findFirst();

        if (result.isEmpty()) {
            throw new NotFoundException("Exercise with publicVisibility and id '333' couldn't be found in the test data");
        }

        Assertions.assertFalse(
            result.get().getSupportedActions().stream().anyMatch(action -> action.getPlugin().equals("Artemis Sharing Connector"))
        );
    }

    @Test
    @WithAnonymousUser
    public void testSearchAll() throws IOException {
        SearchResultsDTO searchResults = searchService.searchAllPublicResults();

        Assert.assertTrue("At least one hit?", searchResults.getSearchResult().size() >= 1);
        Assert.assertEquals("We start at 0", 0, searchResults.getPageStartIndex());
        LOGGER.info("found {} hits for search all", searchResults.getHitCount());
    }
}
